﻿using System;
using System.Collections.Generic;
using System.Drawing;
using System.IO;
using System.Text;
using System.Xml;

namespace FontCombine
{
	/// <summary> グリフのビットマップを敷き詰め、2^nのビットマップフォントを作成するクラス </summary>
    public class Combinner : IDisposable
    {
        List<Glyph> GlyphList = new List<Glyph>();
        public Bitmap texture;
        public int BaseLine;
        public Char DefaultCharacter;
        public int LineHeight;
        public int Spacing;


        public Combinner(String GlyphXMLPath)
        {
            var beforeDirectory = Directory.GetCurrentDirectory();
            Directory.SetCurrentDirectory(Path.GetDirectoryName(GlyphXMLPath));
            XmlDocument XD = new XmlDocument();
            XD.Load(Path.GetFileName(GlyphXMLPath));
            var root = XD.FirstChild;

            this.BaseLine = int.Parse(root.Attributes["mBaseLine"].Value);
            this.DefaultCharacter = root.Attributes["mDefaultCharacter"].Value[0];
            this.LineHeight = int.Parse(root.Attributes["mLineHeight"].Value);
            this.Spacing = int.Parse(root.Attributes["mSpacing"].Value);

            foreach (XmlElement item in XD.GetElementsByTagName("Glyph"))
            {
                GlyphList.Add(new Glyph(item));
            }
            Directory.SetCurrentDirectory(beforeDirectory);
        }

        /// <summary>めっちゃ適当にテクスチャサイズを求めてどうにかする
        /// 　ただしテクスチャサイズは2^n x 2^nになるようにする
        /// </summary>
        public void AnalogyTextureSize1()
        {
            int sumSurface = 0;
            foreach (var item in this.GlyphList)
            {
                sumSurface += item.Size.Width * item.Size.Height;
            }
            var length = Math.Sqrt(sumSurface);
            var normalizedlength = Math.Pow(2, Math.Floor(Math.Log(length, 2)));

			if (texture != null) texture.Dispose();
            texture = new Bitmap((int)normalizedlength, (int)normalizedlength);
        }

        /// <summary>グリフを敷き詰めて1枚のビットマップフォントを作成する
        ///     配置情報はGlyphsList[x].Originを設定することで行う。
        /// </summary>
        public void PlaceGlyphs1()
        {
            if (texture == null)
                throw new Exception("テクスチャサイズが確定しないままグリフを配置しようとしました。この関数を呼ぶ前にAnalogyTextureSizeを呼ぶ必要があります");
            using (Graphics g = Graphics.FromImage(texture))
            {
                GlyphList.Sort();
                //LinkedList<>に仕様か迷う。Remove(item)のための走査コストとか。
                List<Glyph> toolbox = new List<Glyph>(GlyphList);
                int rowTop = 0;

                //int rowHeight = 0;
                    
                while (toolbox.Count > 0)
                {
                
                    List<Glyph> palette = new List<Glyph>();
                    int rowHeight = 0;
                    int left = 0;
                    
                    #region 仮設置
                    //　降順走査はコンテンツの削除以外に、
                    //ソート済みのコンテンツをでかい方から拾いたいため
                    for (int i = toolbox.Count-1; i >= 0; i--)
                    {
                        var item = toolbox[i];
                        if (item.Size.Width + left < texture.Width)
                        {
                            left += item.Size.Width;
                            rowHeight = Math.Max(rowHeight, item.Size.Height);
                            palette.Add(item);
                            toolbox.RemoveAt(i);
                        }
                    }
                    #endregion

                    
                    if (texture.Height < rowTop+ rowHeight)
                    {
                        // 設置できないのでテクスチャサイズを倍にしてやり直し
                        var bmp = new Bitmap(texture.Width * 2, texture.Height * 2);
						texture.Dispose();
						texture = bmp;
						PlaceGlyphs1();
                        return;
                    }



                    left = 0; 
                    #region 実設置
                    foreach (var item in palette)
                    {
                        var top = rowTop + rowHeight - item.Size.Height;
                        item.Origin = new Point(left,top); // new Point(left,top);

                        g.DrawImage(item.image,
                            new Rectangle(new Point(left,top), item.Size),
                            new Rectangle(new Point(0, 0), item.Size), GraphicsUnit.Pixel);
						//g.DrawRectangle(Pens.White, new Rectangle(new Point(left, top), item.Size));
                        //item.Size = new Size(item.Size.Width, rowHeight);
                        left += item.Size.Width;
                    }
                    #endregion
                    rowTop += rowHeight ;
                }

            }

        }

        public void SavePackageXML(String filename)
        {
            using (XmlTextWriter writter = new XmlTextWriter(filename, Encoding.UTF8))
            {
                writter.WriteStartElement("BitmapFont");
                writter.WriteAttributeString("mBaseLine", this.BaseLine.ToString());
                writter.WriteAttributeString("mDefaultCharacter", this.DefaultCharacter.ToString());
                writter.WriteAttributeString("mLineHeight", this.LineHeight.ToString());
                writter.WriteAttributeString("mSpacing", this.Spacing.ToString());

                foreach (var item in GlyphList)
                {
                    item.WriteGlyphAtPackageXML(writter);
                }

                foreach (var item in GlyphList)
                {
                    item.WriteKerningAtPackageXML(writter);
                }

                using (MemoryStream ms = new MemoryStream())
                {
                    this.texture.Save(ms, System.Drawing.Imaging.ImageFormat.Png);
                    ms.Seek(0, SeekOrigin.Begin);
                    Byte[] texBuf = ms.ToArray();
                    writter.WriteStartElement("Textrue");
                    writter.WriteBase64(texBuf, 0, texBuf.Length);
                    writter.WriteEndElement();
                }
                
            }
        }


        #region IDisposable メンバ

        void IDisposable.Dispose()
        {
            GlyphList.ForEach(item => { item.image.Dispose(); });
        }

        #endregion
    }

    // PolygonHeadのGlyphとは微妙に違う
    public class Glyph:IComparable<Glyph>
    {
        public Bitmap image;
        public int AdvanceWidth;
        public Char Character;
        public Boolean ForceWhite;
        public int LeftSideBearing;
        public Point Origin;
        public Size Size;
//        public Size Size { get {return image.Size;} }
        public Dictionary<Char, int> Kerning;

        public Glyph(XmlElement node)
        {
            this.Character = node.GetAttribute("Character")[0];
            this.AdvanceWidth = int.Parse(node.GetAttribute("AdvanceWidth"));
            this.ForceWhite = Boolean.Parse(node.GetAttribute("ForceWhite"));
            this.LeftSideBearing = int.Parse(node.GetAttribute("LeftSideBearing"));
            var filename = node.GetAttribute("src");
            using (Bitmap bmp = new Bitmap(filename))
            {
                image = new Bitmap(bmp);
            }

            //kerning
            Kerning = new Dictionary<char, int>();
            foreach (XmlElement subnode in node.GetElementsByTagName("Kerning"))
            {
                Char c = subnode.GetAttribute("char")[0];
                int value =int.Parse(subnode.GetAttribute("value"));
            }

            this.Size = image.Size;
        }

        public void WriteGlyphAtPackageXML(XmlWriter writter)
        {
            writter.WriteStartElement("mGlyphsItem");
            writter.WriteAttributeString("AdvanceWidth", this.AdvanceWidth.ToString());
            writter.WriteAttributeString("Character", this.Character.ToString());
            writter.WriteAttributeString("ForceWhite", this.ForceWhite.ToString());
            writter.WriteAttributeString("LeftSideBearing", this.LeftSideBearing.ToString());
            writter.WriteAttributeString("OriginX", this.Origin.X.ToString());
            writter.WriteAttributeString("OriginY", this.Origin.Y.ToString());
            writter.WriteAttributeString("SizeX", this.Size.Width.ToString());
            writter.WriteAttributeString("SizeY", this.Size.Height.ToString());
            writter.WriteEndElement();
        }

        public void WriteKerningAtPackageXML(XmlWriter writter)
        {
            if (Kerning.Count <= 0) return;
            
            writter.WriteStartElement("mKerningItem");

            writter.WriteAttributeString("char", this.Character.ToString());
            foreach (var subitem in this.Kerning)
            {
                writter.WriteStartElement("kerning");
                writter.WriteAttributeString("char", subitem.Key.ToString());
                writter.WriteAttributeString("value", subitem.Value.ToString());
                writter.WriteEndElement();
            }
            writter.WriteEndElement();
        }

        #region IComparable メンバ

        
        #endregion

        #region IComparable<Glyph> メンバ

        int IComparable<Glyph>.CompareTo(Glyph other)
        {
            int me = this != null ? this.Size.Width + this.Size.Height * 100 : 0;
            int you = other != null ? other.Size.Width + other.Size.Height * 100 : 0; 

            return (me - you);
        }

        #endregion
    }
}
