﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Xml;
using System.IO;
using System.Drawing;
using System.Xml.Serialization;
using TextRenderer = System.Windows.Forms.TextRenderer;
using TextFormatFlags = System.Windows.Forms.TextFormatFlags;

namespace FontRasterizer
{
    public delegate Bitmap RasterizeChar(String c,Font f);

    public class FontRasterizer
    {
        public Color ForeColor = Color.White;
        public String OutputDirectory = "";
        public String TargetStrings = "";
        public String Prefix = "";
        public int Method = 0;
        public String FontFamilyName = FontFamily.GenericSansSerif.Name;
        public Double FontSize = 24.0;
        public FontStyle FontStyle = FontStyle.Regular;
		public Boolean OverrideOldChar = true;


        public Font GetFont()
        {
			return new Font(this.FontFamilyName, (float)Math.Abs(this.FontSize), this.FontStyle);
        }


        public void SaveAsProject(String filename)
        {
            XmlSerializer xs = new XmlSerializer(typeof(FontRasterizer));
            using (var strm = File.OpenWrite(filename))
            {
                xs.Serialize(strm,this);
            }
        }

        public static FontRasterizer LoadAsProject(String filename)
        {
            XmlSerializer xs = new XmlSerializer(typeof(FontRasterizer));
            using (var strm = File.OpenRead(filename))
            {
                var result = xs.Deserialize(strm) as FontRasterizer;
                result.ForeColor = Color.White; //なぜか保存されない
                return result;
            }            
        }

        public FontRasterizer() { }


        public void Rasterize()
        {
            switch (this.Method)
            {
                case 1:
                    Rasterize(this.RastCharWithDrawString);
                    break;
                case 2:
                    Rasterize(this.RastCharWithTextRender);
                    break;
                case 3:
                    Rasterize(this.RastCharWithTextRenderShrink);
                    break;
                case 4:
                    Rasterize(this.RastCharWithDrawStringShrink);
                    break;
				case 5:
					Rasterize(this.RastCharWithDrawStringAndMCR);
					break;
				case 6:
					Rasterize(this.RastCharWithDrawStringAndMCRAndSpaceTrick);
					break;
				default:
					Rasterize(this.RastCharWithDrawStringAndMCRAndSpaceTrick);
                    break;

            }
        }

		private void AutoFontSize()
		{
			const String sampleChar = "a";
			Size targetSize = new Size();
			Double fontSize = 15; // default;
			int recalcTime = 5; 

			// ## Get targetSize ####################################
			var XD = new System.Xml.XmlDocument();
			XD.Load(this.OutputDirectory + "\\glyphs.xml");

			String sc = sampleChar == "\"" ? "'\"'" : "\"" + sampleChar + "\"";
			foreach (XmlElement item in XD.SelectNodes(String.Format("//Glyph[@Character={0}]", sc)))
			{
				String bmpFilename = item.GetAttribute("src");
				using (Bitmap bmp = new Bitmap(this.OutputDirectory + "\\" + bmpFilename))
				{
					targetSize = bmp.Size;
				}
			}

			// ## Calc fontSize #####################################
			for (int i = 0; i < recalcTime; i++)
			{
				Font tempFont = new Font(this.FontFamilyName,(float)fontSize,this.FontStyle);
				SizeF realSize = FontSizeMesure.MesureSize2(tempFont, sampleChar);
				fontSize = fontSize * (targetSize.Width / realSize.Width);
			}
			this.FontSize = fontSize ;
		}

        public void Rasterize(RasterizeChar rasterizeChar)
        {
			if (this.FontSize <= 0) AutoFontSize();
			Font font = this.GetFont();

            List<Char> outputedChar = new List<char>();

            XmlDocument XD = new XmlDocument();
            var beforeCurrentDirectory = Directory.GetCurrentDirectory();
            Directory.SetCurrentDirectory(OutputDirectory);
            if (File.Exists("glyphs.xml"))
                XD.Load("glyphs.xml");


			if (!OverrideOldChar)
			{
				foreach (XmlElement item in XD.GetElementsByTagName("Glyph"))
				{
					Char c = item.GetAttribute("Character")[0];
					outputedChar.Add(c);
				}
			}


            int charCount = 0;
            foreach (var c in TargetStrings.ToCharArray())
            {
                if (c == '\r' || c == '\n') continue;
                if (outputedChar.Contains(c)) continue;

                outputedChar.Add(c);
                using (var bmp = rasterizeChar(c.ToString(), font))
                {
                    var charFilename = String.Format("{0}_{1:D4}.png", this.Prefix, charCount);
                    bmp.Save(charFilename);

					String sc = c == '"' ? "'\"'" : "\"" + c.ToString() + "\"";
					foreach (XmlElement item in XD.SelectNodes(String.Format("//Glyph[@Character={0}]", sc)))
					{
						item.ParentNode.RemoveChild(item);
					}

                    var elem = XD.CreateElement("Glyph");
                    elem.SetAttribute("Character", c.ToString());
                    elem.SetAttribute("AdvanceWidth", bmp.Width.ToString());
                    elem.SetAttribute("ForceWhite", "False"); //??
                    elem.SetAttribute("LeftSideBearing", "0");
                    elem.SetAttribute("src", charFilename);

					if (c == ' ') elem.SetAttribute("AdvanceWidth", "-3");
                    XD.FirstChild.AppendChild(elem);

                    charCount += 1;
                }
            }
            XD.Save("glyphs.xml");

            Directory.SetCurrentDirectory(beforeCurrentDirectory);
        }

        private Bitmap RastCharWithTextRenderShrink(String c, Font f)
        {
            TextFormatFlags tfFlag = TextFormatFlags.NoPadding | TextFormatFlags.NoPrefix;
            var size = TextRenderer.MeasureText(c, f, new Size(1000, 1000), tfFlag);

            using (var prerender = new Bitmap(size.Width, size.Height))
            {
                using (var g = Graphics.FromImage(prerender))
                {
                    g.TextRenderingHint = System.Drawing.Text.TextRenderingHint.SingleBitPerPixel;
                    TextRenderer.DrawText(g, c, f, new Point(0, 0), this.ForeColor, tfFlag);
                }
                var realSize = prerender.Size;

                #region 余白部分検出
                for (int x = realSize.Width - 1; x >= 0; x--)
                {
                    Boolean isBlank = true;
                    for (int y = 0; y < realSize.Height; y++)
                    {
                        isBlank = isBlank && prerender.GetPixel(x, y).A == 0;
                    }
                    if (isBlank)
                    {
                        realSize.Width = x;
                    }
                    else
                    {
                        break;
                    }
                }

                for (int y = realSize.Height - 1; y >= 0; y--)
                {
                    Boolean isBlank = false;
                    for (int x = 0; x < realSize.Width; x++)
                    {
                        isBlank = isBlank && prerender.GetPixel(x, y).A == 0;
                    }
                    if (isBlank)
                    {
                        realSize.Height = y;
                    }
                    else
                    {
                        break;
                    }
                }

                #endregion

                if (realSize.Width == 0 || realSize.Height == 0)
                {
                    //全角のスペース"　"のたぐいの可能性が高い
                    realSize = prerender.Size;
                }

                var result = new Bitmap(realSize.Width, realSize.Height);
                using (var g = Graphics.FromImage(result))
                {
                    g.DrawImage(
                        prerender,
                        new Rectangle(0, 0, realSize.Width, realSize.Height),
                        new Rectangle(0, 0, realSize.Width, realSize.Height),
                        GraphicsUnit.Pixel);
                }
                return result;
            }
        }

        private Bitmap RastCharWithTextRender(String c, Font f)
        {
            TextFormatFlags tfFlag = TextFormatFlags.NoPadding | TextFormatFlags.NoPrefix;

            var size = TextRenderer.MeasureText(c, f,new Size(1000,1000), tfFlag);
            var result = new Bitmap(size.Width, size.Height);
            using (var g = Graphics.FromImage(result))
            {
                g.TextRenderingHint = System.Drawing.Text.TextRenderingHint.AntiAliasGridFit;
                TextRenderer.DrawText(g, c, f, new Point(0, 0), this.ForeColor,tfFlag);
            }
            return result;
        }

        private Bitmap RastCharWithDrawString(String c,Font f)
        {
            var size = FontSizeMesure.MesureSize(f, c);
            var result = new Bitmap(size.Width, size.Height);
            using (var g = Graphics.FromImage(  result))
            {
                g.TextRenderingHint = System.Drawing.Text.TextRenderingHint.AntiAliasGridFit;
                g.DrawString(c, f, new SolidBrush(this.ForeColor), new PointF(0, 0));
            }
            return result;
        }

        private Bitmap RastCharWithDrawStringShrink(String c, Font f)
        {
            using (Bitmap basebmp = RastCharWithDrawString(c, f))
            {
                var cropRegion = FindCropArea(basebmp);
                Bitmap result = new Bitmap(cropRegion.Width, cropRegion.Height);
                using (var g = Graphics.FromImage(result))
                {
                    g.DrawImage(
                        basebmp,
                        new Rectangle(new Point(0, 0), cropRegion.Size),
                        cropRegion,
                        GraphicsUnit.Pixel);
                }

                return result;
            }
        }

        private Bitmap RastCharWithDrawStringAndMCR(String c, Font f)
        {
            RectangleF realRect;
            using (Graphics g = Graphics.FromImage(FontSizeMesure.DummyBitmap))
            {
                StringFormat sformat = new StringFormat();
                sformat.SetMeasurableCharacterRanges(new[] { new CharacterRange(0, 1) });
                var region = g.MeasureCharacterRanges(c, f, new RectangleF(0, 0, 1000,1000), sformat)[0];
                realRect = region.GetBounds(g);
            }
            if (realRect.Width <= 0 || realRect.Height <= 0)
            {
                return this.RastCharWithDrawString(c, f);
            }


            Bitmap result = new Bitmap((int)realRect.Width, (int)realRect.Height);
            using (Graphics g = Graphics.FromImage(result))
            {
                g.TextRenderingHint = System.Drawing.Text.TextRenderingHint.AntiAliasGridFit;
                g.DrawString(c, f, new SolidBrush(this.ForeColor),
                    new PointF(-realRect.X, -realRect.Y));
            }

            return result;
        }

		private Bitmap RastCharWithDrawStringAndMCRAndSpaceTrick(String c, Font f)
		{

			RectangleF realRect;
			using (Graphics g = Graphics.FromImage(FontSizeMesure.DummyBitmap))
			{
				StringFormat sformat = new StringFormat();
				sformat.SetMeasurableCharacterRanges(new[] { new CharacterRange(0, 1) });
				var region = g.MeasureCharacterRanges(c, f, new RectangleF(0, 0, 1000, 1000), sformat)[0];
				realRect = region.GetBounds(g);
			}
			if (realRect.Width <= 0 || realRect.Height <= 0)
			{
				return this.RastCharWithDrawString(c, f);
			}


			Bitmap result = new Bitmap((int)realRect.Width, (int)realRect.Height);
			using (Graphics g = Graphics.FromImage(result))
			{
				g.TextRenderingHint = System.Drawing.Text.TextRenderingHint.AntiAliasGridFit;
				g.DrawString(c, f, new SolidBrush(this.ForeColor),
					new PointF(-realRect.X, -realRect.Y));
			}

			return result;
		}

        private Rectangle FindCropArea(Bitmap bmp)
        {
            int top = 0;
            int left = 0;
            int right = 0;
            int bottom = 0;
            Boolean flagTop = false;
            Boolean flagLeft = false;
            Boolean flagRight = false;
            Boolean flagBottom = false;

            var w = bmp.Width;
            var h = bmp.Height;
            var sqr = w * h;

            for (int i = 0; i < sqr; i++)
            {
                int x1 = i % w;
                int y1 = i / w;

                int x2 = i / h;
                int y2 = i % h;

                int x3 = w - x2-1;
                int y3 = y2;

                int x4 = x1;
                int y4 = h - y1-1;

                // Top
                if (bmp.GetPixel(x1, y1).A != 0 && !flagTop)
                {
                    top = y1;
                    flagTop = true;
                }
                // left
                if (bmp.GetPixel(x2, y2).A != 0 && !flagLeft)
                {
                    left = x2;
                    flagLeft = true;
                }
                // right
                if (bmp.GetPixel(x3, y3).A != 0 && !flagRight)
                {
                    right = x3;
                    flagRight = true;
                }
                // bottom
                if (bmp.GetPixel(x4, y4).A != 0 && !flagBottom)
                {
                    bottom = y4;
                    flagBottom = true;
                }

                // break
                if (flagBottom && flagLeft && flagRight && flagTop)
                {
                    System.Diagnostics.Debug.WriteLine(String.Format("{0} ,{1} ,{2} ,{3} ({4}x{5})", left, top, right, bottom,right-left,bottom-top));
                    return Rectangle.FromLTRB(left, top, right+1, bottom+1);
                }
            }
            return new Rectangle(0, 0, w, h); // スペースなどの場合    
            

            #region old code Top
            /*
            for (int y = 0; y < bmp.Height; y++)
            {
                for (int x = 0; x < bmp.Width; x++)
                {
                    if (bmp.GetPixel(x, y).A != 0)
                    {
                        realRect.Y = y;
                        y = bmp.Height; // break for y
                        break;
                    }
                }
            }
            #endregion
            // 空白用
            if (realRect.Y == -1)
                return new Rectangle(0, 0, bmp.Width, bmp.Height);

            #region Left
            for (int x = 0; x < bmp.Width; x++)
            {
                for (int y = realRect.Y; y < bmp.Height; y++)
                {
                    if (bmp.GetPixel(x, y).A != 0)
                    {
                        realRect.X = x;
                        x = bmp.Width; // break for x;
                        break;
                    }
                }
            }

            #endregion

            #region Bottom
            for (int y = bmp.Width-1; y >= realRect.Y; y--)
            {
                for (int x = realRect.X; x < bmp.Width; x++)
                {
                    if (bmp.GetPixel(x, y).A != 0)
                    {
                        realRect.Height = y - realRect.Y;
                        y = -1;  // break for
                        break;
                    }
                }
            }

            #endregion

            #region Right
            for (int x = bmp.Width-1; x >= realRect.X; x--)
            {
                for (int y = realRect.Y; y < realRect.Height; y++)
                {
                    if (bmp.GetPixel(x, y).A != 0)
                    {
                        realRect.Width = x - realRect.X;
                        x = -1;  // break for
                        break;
                    }
                }
            }
            */
            #endregion


        }

        #region IDisposable メンバ


        #endregion
    }

    public static class FontSizeMesure
    {
        static Bitmap singleton ;
        static Object singletonLock = new Object();
        
        public static Bitmap DummyBitmap
        { 
            get
            {
                lock(singletonLock)
                {
                    if (singleton == null) 
                    {
                        singleton = new Bitmap(1,1);
                    }
                    return singleton;
                }
            
            }
        }
        
        public static Size MesureSize(Font font,String s)
        {
            using (Graphics g = Graphics.FromImage(DummyBitmap))
            {
                return g.MeasureString(s,font).ToSize();
            }
        }

		public static SizeF MesureSize2(Font font, String s)
		{
			using (Graphics g = Graphics.FromImage(FontSizeMesure.DummyBitmap))
			{
				StringFormat sformat = new StringFormat();
				sformat.SetMeasurableCharacterRanges(new[] { new CharacterRange(0, 1) });
				var region = g.MeasureCharacterRanges(s, font, new RectangleF(0, 0, 1000, 1000), sformat)[0];
				return region.GetBounds(g).Size;
			}
		}

    }
}
